import csv
import os
import re
from contextlib import contextmanager

import jinja2schema

from edam import get_logger
from edam.reader.regular_expressions import template_file_header, \
    for_loop_variables, var_for_line, start_if, end_if, resample_function
from edam.utilities.exceptions import ErrorWithTemplate

logger = get_logger('edam.reader.models.template')


class Template:
    filename: str
    path: str
    variables: str

    def __init__(self, path=None):
        self.path = path

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def header(self) -> str:
        """Gets the header of template based on a regex

        :rtype: str
        :return: Header as string
        """
        regex_header_from_template_file = re.compile(template_file_header)
        with read_template(self) as template_file_object:
            template_contents = template_file_object.read()

        matches = re.findall(regex_header_from_template_file, template_contents)
        if matches:
            header = matches[0][0].strip("\r\n")
            return header
        logger.warning(f"{self.filename} does not have header")

    @property
    def header_line(self):
        with read_template(self) as f:
            for line_number, line in enumerate(f, 0):
                if self.header in line:
                    return line_number

    @property
    def dataframe_header(self):
        return list(filter(lambda x: x != '',
                           self.stripped_contents.split(self.delimiter)))

    @property
    def used_columns(self):
        columns = list(filter(lambda x: x != '',
                              self.stripped_contents.split(self.delimiter)))
        used_columns = list()
        for column in columns:
            variable = jinja2schema.infer(column)
            if len(variable.keys()) > 0:
                for key, value in variable.items():
                    attribute = list(value.keys()).pop()
                    if attribute == "value":
                        used_columns.append(f"{key}")
                    else:
                        used_columns.append(f"{key}.{attribute}")
            else:
                used_columns.append(column)
        return used_columns

    @property
    def variables(self) -> [str]:
        variables = jinja2schema.infer(self.cleaned_contents)
        return list(variables.keys())

    @property
    def observable_ids(self) -> [str]:
        """
           This function parses a template file and returns
           the variables for the template in "for loop" (i.e. observable_id's).
           It returns a list with the variables

           This function is useful for "viewing" purposes.
           A user can submit a query to the web portal and find
           all available templates along with the corresponding observable_id's

           :rtype: [str]
           :return: List of observable IDs
        """
        with read_template(self) as template_file_object:
            template_contents = template_file_object.read()
        matches = re.findall(for_loop_variables, template_contents)
        if matches:
            template_observables = matches[0][0]  # type: str
            self.resampled = matches[0][1]
            template_observables_as_list = list(
                map(lambda observable: observable.rstrip().lstrip(),
                    template_observables.split(',')))

            return template_observables_as_list
        raise ErrorWithTemplate(f"I couldn't extract variables from "
                                f"{self.filename} located at {self.path}")

    @property
    def resampled(self) -> dict:
        """
        """
        return self._resampled

    @resampled.setter
    def resampled(self, value):
        matches = re.findall(resample_function, value)
        dictionary = {}
        if matches:
            try:
                rule = matches[0][0]
            except:
                rule = None
            try:
                how = matches[0][1]
            except:
                rule = None
            dictionary['rule'] = rule
            dictionary['how'] = how
            self._resampled = dictionary
        else:
            self._resampled = None


    @property
    def preamble(self) -> str:
        """
        Gets the template's preamble text (if applicable)
        :return:
        """
        with read_template(self) as template_file_object:
            template_contents = template_file_object.read()

        preamble, _, _ = template_contents.partition(self.header)
        preamble = preamble.rstrip('\n\r')
        if preamble == "":
            return None
        return preamble

    @property
    def stripped_contents(self) -> str:
        with read_template(self) as template_file_object:
            template_contents = template_file_object.read()
        matches = re.findall(var_for_line, template_contents)
        try:
            return matches.pop().lstrip('\n\r').rstrip('\n\r')
        except Exception as exc:
            raise ErrorWithTemplate(exc)

    @property
    def cleaned_contents(self):
        temp = re.sub(start_if, '', self.stripped_contents)
        temp = re.sub(end_if, '', temp).lstrip('\r\n').rstrip('\r\n')
        return temp

    @property
    def delimiter(self) -> str:
        dialect = csv.Sniffer().sniff(self.cleaned_contents)
        return str(dialect.delimiter)

    @property
    def name(self) -> str:
        return self.filename.split('.tmpl')[0]

    def to_dict(self):
        temp = {}
        properties = ['delimiter', 'filename', 'header', 'header_line',
                      'variables']
        for prop in properties:
            try:
                temp[prop] = self.__getattribute__(prop)
            except Exception as e:
                logger.exception("Exception: ")
                temp[prop] = "ERROR"
        return temp

    @property
    def contents(self):
        with read_template(self) as template_file_object:
            return template_file_object.read()

    def __repr__(self):
        return f"{self.filename} located at {self.path}"


@contextmanager
def read_template(template: Template):
    """
    Returns the templates file object
    :rtype: typing.TextIO
    :param template:
    """
    f = open(template.path, 'r')
    yield f
    f.close()



class Error(Exception):
    """Base class for other exceptions"""
    pass


class TemplateDoesNotExist(Error):
    """Raised when template does not exist"""
    pass


class MetadataFileDoesNotExist(Error):
    """Raised when template does not exist"""
    pass


class ErrorWithTemplate(Error):
    """Raised when template does not exist"""
    pass


class FileDoesNotExist(Error):
    """Raised when file does not exist"""
    pass


class InputParameterDoesNotExist(Error):
    """Raised when InputParameter does not exist"""
    pass


class UrlInputParameterDoesNotExist(Error):
    """Raised when UrlInputParameter does not exist"""
    pass


class TemplateInputHeaderMismatch(Error):
    """Raised when Template and Input file headers do not match each other"""
    pass


class InvalidUsage(Exception):
    status_code = 400

    def __init__(self, message, status_code=None, payload=None):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self.payload = payload

    def to_dict(self):
        rv = dict(self.payload or ())
        rv['message'] = self.message
        return rv



from edam.reader.resolvers.resolver import Resolver


class FileResolver(Resolver):

    @property
    def content_as_list(self):
        return self._content_as_list

    @content_as_list.setter
    def content_as_list(self, file_uri):
        with open(file_uri, 'r') as file:
            self._content_as_list = list(
                map(lambda line: line.strip('\n\r'), file.readlines()))



class Measurement(object):
    def __init__(self, value, timestamp=None, observable=None,
                 uom=None, station=None, junction=None):
        self.value = value
        self.timestamp = timestamp
        self.observable = observable
        self.uom = uom
        self.station = station
        self.junction = junction

    def __repr__(self):
        return f"{self.value}"



import functools
import time

from edam import get_logger

logger = get_logger(__name__)


def timer(func):
    """Print the runtime of the decorated function"""

    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()
        value = func(*args, **kwargs)
        end_time = time.perf_counter()
        run_time = end_time - start_time
        logger.debug(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        return value

    return wrapper_timer



import copy
import json
from typing import TYPE_CHECKING

import pandas as pd

from edam.reader.database_handler import update_object, exists

if TYPE_CHECKING:
    from edam.reader.models.template import Template
    from edam.reader.models.station import Station


def resample(station: "Station", rule, how=None):
    dataframe = station.data
    rule = rule.lstrip('"').lstrip("'").rstrip('"').rstrip("'")
    observables_list = list(dataframe)
    observables_list.remove('timestamp')
    available_operations = [
        'bfill',
        'max',
        'median',
        'sum',
        'min',
        'interpolate',
        'ffill',
        'mean']

    try:
        for observable in observables_list:
            dataframe[observable] = dataframe[observable].apply(
                lambda x: float(x))

    except Exception as e:
        # logger.exception("Exception")
        pass
    dataframe['timestamp'] = pd.to_datetime(dataframe['timestamp'])
    resampled = dataframe.resample(rule=rule, on='timestamp')

    if how is None:
        resampled = resampled.mean()
        resampled = resampled.round(3)
        resampled = resampled.fillna('---')
    else:
        how = how.lstrip('"').lstrip("'").rstrip('"').rstrip("'")
        if how in available_operations:
            resampled = getattr(resampled, how)()
            resampled = resampled.round(3)

    resampled['timestamp'] = resampled.index
    for observable in list(resampled):
        resampled[observable] = resampled[observable].apply(
            lambda x: str(x))
    return resampled


def update_existing(obj, new_values, logger):
    for key, value in new_values.items():
        try:
            if key in ['tags', 'qualifiers']:
                temp = getattr(obj, key)  # type: dict
                try:
                    value.update(temp)
                except ValueError as exception:
                    logger.warning(f"{exception}")
            setattr(obj, key, value)
        except AttributeError:
            logger.warning(f"{key} does not exist")
    if exists(obj) is None:
        update_object(obj)


def as_dict(obj):
    temp_dict = copy.deepcopy(obj.__dict__)  # type: dict
    temp_dict.pop("_sa_instance_state")
    for key in list(temp_dict.keys()):
        if key.startswith('_'):
            temp_dict[key.lstrip('_')] = json.loads(temp_dict[key])
            temp_dict.pop(key)

    return temp_dict



import json
from datetime import datetime

import pandas as pd
from sqlalchemy.dialects import sqlite
from sqlalchemy.ext.hybrid import hybrid_property

from sqlalchemy import Column, Integer, String, Boolean, Float
from sqlalchemy.orm import relationship

from edam import get_logger
from edam.reader.base import Base, session, engine
from edam.reader.models.measurement import Measurement
from edam.reader.models.observation import Observation
from edam.reader.models.utilities import update_existing, as_dict, resample
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from edam.reader.models.template import Template

logger = get_logger('edam.reader.models.station')


class Station(Base):
    """
    Represents a Station.

    This ORM class represents a station along with its metadata.
    According to the EDAM conceptual model, a station can be stationary or
    moving.

    Attributes:
        name: The name of the station
        mobile: Boolean value to define whether it's a stationary or moving
            station
        location: string attribute which describes the location where the
            station is situated
        latitude: float attribute which represent the station's latitude
        longitude: float attribute which represent the station's longitude
        region: string attribute which describes the region where the
            station is situated. TODO: Obsolete??
        license: string attribute which represents the license of the data
            that the station holds
        url: string attribute of the station's url (if applicable)
        tags: dict attribute which represents any other tags which are not
            explicitly defined
        qualifiers: dict attribute which represents any qualifiers for the
            station's data. A standard qualifier is `missing_data`. For the
            rest, the `key` of this attribute is the qualifier (e.g. `*`) and
            the `value` the qualifier (e.g. `estimated`)
    """
    __tablename__ = "Station"
    id = Column(Integer, primary_key=True)
    name = Column(String(80))
    mobile = Column(Boolean)
    location = Column(String(200))
    latitude = Column(Float)
    longitude = Column(Float)
    region = Column(String(200))
    license = Column(String(100))
    url = Column(String(100))
    _tags = Column('tags', String(500))
    _qualifiers = Column('qualifiers', String(500))

    junctions = relationship("Junction", back_populates="station")

    def update(self, new_values: dict):
        update_existing(self, new_values, logger)

    def __init__(self,
                 name: str = None, mobile: bool = False,
                 location: str = None, latitude: float = None,
                 longitude: float = None, region: str = None,
                 license: str = None, url: str = None,
                 qualifiers: dict = None, tags: dict = None):
        if tags is None:
            tags = dict()
        self.name = name
        self.mobile = mobile
        self.location = location
        self.latitude = latitude
        self.longitude = longitude
        self.region = region
        self.license = license
        self.url = url

        self.qualifiers = qualifiers
        self.tags = tags

    @hybrid_property
    def tags(self):
        return json.loads(self._tags)

    @tags.setter
    def tags(self, value):
        self._tags = json.dumps(value)

    @hybrid_property
    def qualifiers(self):
        return json.loads(self._qualifiers)

    @qualifiers.setter
    def qualifiers(self, value):
        if value is None:
            self._qualifiers = json.dumps({})
        else:
            self._qualifiers = json.dumps(value)

    @property
    def missing_data(self):
        try:
            return self.qualifiers['missing_data']
        except KeyError:
            return ''
        except TypeError:
            return ''

    @property
    def observable_ids(self) -> [str]:
        return [junction.observable.observable_id for junction in
                self.junctions]

    @property
    def junctions_mapping(self) -> dict:
        return {junction.id: junction.observable.observable_id for junction in
                self.junctions}

    @property
    def data(self) -> pd.DataFrame:
        dataframes = []
        for junction in self.junctions_mapping.keys():
            sql = session.query(Observation).filter(
                Observation.junction_id == junction)
            sql_literal = str(sql.statement.compile(dialect=sqlite.dialect(),
                                                    compile_kwargs={
                                                        "literal_binds": True}))

            df = pd.read_sql_query(sql_literal, engine)

            df.rename(columns={"value": self.junctions_mapping[junction]},
                      inplace=True)
            if dataframes:
                df.drop(['id', 'junction_id', 'timestamp'], axis=1,
                        inplace=True)
            else:
                df.drop(['id', 'junction_id'], axis=1, inplace=True)
            dataframes.append(df)
        if dataframes:
            dataframe = pd.concat(dataframes, join='inner', axis=1).fillna(
                "empty")  # type: pd.DataFrame
            dataframe.set_index(keys=["timestamp"], drop=False, inplace=True)
            return dataframe
        logger.warning(f"{self.name} does not have any data associated")
        return None

    def data_iter(self, template: "Template"):
        data = self.data
        variables = template.observable_ids
        if template.resampled:
            data = resample(self, template.resampled['rule'],
                            template.resampled['how'])
        for row in data.itertuples():
            if variables is not None:
                yield self._prepare_dataframe_generator(row, variables)
            else:
                yield self._prepare_dataframe_generator(row, list(data))

    def resampled(self, rule=None, how=None):
        return self._res

    @property
    def res(self):
        return self._res

    @res.setter
    def res(self, value):
        self._res = value

    def _prepare_dataframe_generator(self, row: tuple, variables) -> []:
        gen = []
        for variable in variables:
            value = row.__getattribute__(variable)
            if variable == "timestamp":
                gen.append(datetime.strptime(value.split('.')[0],
                                             "%Y-%m-%d %H:%M:%S"))
            else:
                if value == "nan":
                    try:
                        gen.append(Measurement(self.qualifiers['missing_data']))
                    except Exception:
                        gen.append(Measurement(value))
                else:
                    gen.append(Measurement(value))
        return gen

    def as_dict(self):
        return as_dict(self)

    def __eq__(self, other):
        """Override the default Equals behavior"""
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        return NotImplemented

    def __ne__(self, other):
        """Define a non-equality test"""
        if isinstance(other, self.__class__):
            return not self.__eq__(other)
        return NotImplemented

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name!r} with id {self.id!r}>'



import json
import os.path

from flask import render_template
from flask import request, make_response, jsonify
from flask_googlemaps import GoogleMaps
from flask_googlemaps import Map

import plotly
import plotly.graph_objs as go

from edam import get_logger
from edam.reader.models.station import Station
from edam.reader.models.template import Template
from edam.utilities.exceptions import InvalidUsage
from edam.viewer.app import app, stations, nocache, templates, \
    render_data
from edam.viewer.app.OgcSos import OgcSos

logger = get_logger('edam.viewer.app.views')

GoogleMaps(app)


@app.route('/maps/')
@app.route('/index/')
@app.route('/')
def mapview():
    markers = list()
    for station in stations():
        if station.latitude is not None and station.longitude is not None:
            observables = ', '.join(
                [junction.observable.name for junction in station.junctions])

            infobox = f'<p><b>Station</b>: ' \
                      f'<a href="{request.url_root}Stations/{station.id}" ' \
                      f'target="_blank">{station.name}</a></p>' \
                      f'<p><b>Observables</b>: {observables}</p>'
            temp_dict = {
                'lat': station.latitude,
                'lng': station.longitude,
                'infobox': infobox}
            markers.append(temp_dict)
    mymap = Map(
        identifier="map-canvas",
        lat=38.045362,
        lng=23.715078,
        zoom=3,
        markers=markers,
        style="width:100%;height:650px;"
    )
    return render_template('maps.html', mymap=mymap)


@app.errorhandler(InvalidUsage)
def handle_invalid_usage(error):
    return render_template('error.html', errorTitle=str(
        error.status_code), errorMessage=error.message)


@app.route('/Stations')
@app.route('/Stations/')
@app.route('/stations')
@app.route('/stations/')
def stat():
    stations_dict = {station.id: station.as_dict() for station in stations()}
    return stations_dict


@app.route('/home')
@app.route('/home/')
def index():
    return render_template('home.html')


@app.route('/Stations/<station>')
@app.route('/stations/<station>')
def specific_station(station):
    stations_dict = {station.id: station.as_dict() for station in stations()}
    return jsonify(stations_dict[int(station)])


@app.route('/templates')
@app.route('/templates/')
def templ():
    templates_dict = {template.name: template.to_dict() for template in
                      templates()}
    return templates_dict


@app.route('/templates/<template>')
def specific_template(template):
    try:
        temp = next(filter(lambda item: item.name == template, templates()))

        response = make_response(temp.contents)
        response.headers['Content-type'] = 'text/plain'
        return response
    except StopIteration:
        raise InvalidUsage(
            f'{template} template does not exist', status_code=410)


@app.route('/data')
@nocache
def get_data():
    station_name = request.args.get('station')
    template_name = request.args.get('template')
    try:
        station = next(filter(lambda item: item.name == station_name,
                              stations()))  # type: Station
    except StopIteration:
        message = f"Station {station_name} does not exist"
        logger.warn(message)
        raise InvalidUsage(message, status_code=410)
    try:
        template = next(
            filter(lambda item: item.name == template_name,
                   templates()))  # type: Template
    except StopIteration:
        message = f"Template {template_name} does not exist"
        logger.warn(message)
        raise InvalidUsage(message, status_code=410)

    data = render_data(template=template,
                       station=station)
    if data is not None:
        if template.resampled:
            station.res = data
            response = make_response(
                render_template(os.path.join('edam', template.filename),
                                station=station))
        else:
            response = make_response(
                render_template(os.path.join('edam', template.filename),
                                chunk=data,
                                station=station))
        response.headers['Content-type'] = 'text/plain'
        response.headers['Cache-Control'] = 'public, max-age=0'

        return response
    else:
        raise InvalidUsage(
            f'{template_name} template and {station.name} station '
            f'are not compatible',
            status_code=410)


@app.route('/about/')
def about():
    return render_template('about.html')


@app.route('/graphs')
def graphs():
    all_stations = stations()  # type: [Station]
    metrics = all_stations[0].observable_ids
    return render_template('graphs.html', stations=all_stations,
                           metrics=metrics)


@app.route('/line', methods=['GET', 'POST'])
def line_graph():
    metrics = request.args['metrics'].split(',')
    st = request.args['stations'].split(',')
    graphJSON = line_plotter(metrics, st)
    data_to_return = dict()
    data_to_return['data_json'] = graphJSON
    data_to_return['layout'] = dict()
    data_to_return['layout']['title'] = f"{','.join(metrics)}"

    return data_to_return


def line_plotter(metrics, st):
    tracers = list()
    for sta in st:
        try:
            station = next(filter(lambda sts: sts.name == sta, stations()))
        except StopIteration:
            return {}
        for metric in metrics:
            df = station.data
            metric_formal = next(
                filter(lambda it: it.observable.observable_id == metric,
                       station.junctions))

            trace = go.Scatter(
                x=df.index,
                y=df[metric],
                mode='lines',
                name=f'{station.name} - {metric_formal.observable.name.capitalize()}'
            )
            tracers.append(trace)
    graphJSON = json.dumps(tracers, cls=plotly.utils.PlotlyJSONEncoder)
    return graphJSON


@app.route('/SensorObservationService/')
def sensor_info():
    return render_template('SOS_Examples.html')


@app.route('/sos/')
@app.route('/Sos/')
@app.route('/sos/')
@app.route('/SOS/')
def sos():
    sos_request = request.args.get('request', '')
    sos_procedure = request.args.get('procedure', '')
    sos_offering = request.args.get('offering', '')
    sos_observed_property = request.args.get('observedProperty', '')
    sos_event_time = request.args.get('eventTime', '')

    sos = OgcSos(
        sos_request,
        sos_procedure,
        sos_offering,
        sos_event_time,
        sos_observed_property)
    sos_response = sos

    response = make_response(
        render_template(
            sos_response.template,
            info=sos.info,
            keywords=sos.keywords,
            stations=sos.stations,
            helpers=sos.metadata,
            procedure=sos.procedure,
            sensor=sos.sensor,
            results=sos.results,
            helper=sos.helper_object))
    response.headers["Content-Type"] = "application/xml"

    return response


if __name__ == "__main__":
    pass



from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import scoped_session, sessionmaker

from edam import database_url

engine = create_engine(database_url, connect_args={'check_same_thread': False})
Session = sessionmaker(bind=engine, autoflush=True, autocommit=False,
                       expire_on_commit=True)
session = Session(future=True)
Base = declarative_base(bind=engine)
Base.query = scoped_session(Session).query_property()



from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from edam.reader.base import Base


class UnitOfMeasurement(Base):
    """
    Represents a unit of measurement.

    This ORM class represents a unit of measurement.

    Attributes:
        name: The name of the unit (e.g. Celcius)
        ontology: An ontology for this unit
        symbol: The symbol for the unit (e.g. C)
    """
    __tablename__ = "UnitOfMeasurement"
    id = Column(Integer, primary_key=True)
    name = Column(String(60))
    ontology = Column(String(160))
    symbol = Column(String(15))

    junctions = relationship("Junction", back_populates="unit")

    def __init__(self, name=None, ontology=None, symbol=None):
        self.name = name
        self.ontology = ontology
        self.symbol = symbol

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name!r} with id {self.id!r}>'



import copy

from edam import get_logger
from edam.reader.base import session

logger = get_logger('edam.reader.logger')


def exists(item):
    try:
        item_dict = copy.deepcopy(item.__dict__)  # type: dict

        item_dict.pop('_sa_instance_state')
        _item = session.query(item.__class__).filter_by(**item_dict).first()
        return _item
    except Exception:
        logger.exception(
            f"Exception while checking whether {item} exists in db")
        return None


def add_item(item):
    """
    Stores or updates an item in the database.

    In case insertion is successful
    it returns the item (complemented with the database id). In case it's not
    it will raise an exception.

    Args:
        item: object to be stored in database

    Returns:
        The item stored in the database (along with the ID)
    """
    try:
        existing_object = exists(item)
        if existing_object is None:
            session.add(item)
            session.commit()
            logger.debug(f"Added {item} in db")
            return item
    except BaseException:
        logger.exception(
            f'Exception when adding {item}')
        session.rollback()
        raise
    else:
        return existing_object


def get_all(item):
    session.flush()
    return session.query(item).all()


def add_items(items: list):
    """
    Adds a list of items in database.

    Calls the `add_item` on multiple items in a list.

    Args:
        items: list with items to be added in database

    Returns:
        The items as added in the database
    """
    unique_items = list(filter(lambda item: exists(item) is None, items))
    if unique_items:
        session.bulk_save_objects(unique_items)
        session.commit()
        session.flush()


def update_object(item):
    session.merge(item)
    session.commit()
    session.flush()



import os

import yaml

from edam.reader.database_handler import add_item
from edam.reader.models.observable import AbstractObservable
from edam.reader.models.sensor import Sensor
from edam.reader.models.station import Station
from edam.reader.models.unit_of_measurement import UnitOfMeasurement


class Metadata:
    def __init__(self, path=None):
        self.path = path

    @property
    def restructured_metadata(self):
        restr_meta = dict()
        observables = self.observables
        sensors = self.sensors
        unit = self.units_of_measurement
        for obs_id in self.observable_ids:
            restr_meta['station'] = self.station
            restr_meta[obs_id] = dict()
            restr_meta[obs_id]['sensor'] = sensors[obs_id]
            restr_meta[obs_id]['observable'] = observables[obs_id]
            restr_meta[obs_id]['unit_of_measurement'] = unit[obs_id]

        return restr_meta

    @property
    def filename(self):
        return os.path.basename(self.path)

    @property
    def contents(self):
        try:
            with open(self.path, 'r') as f:
                return yaml.load(f.read(), Loader=yaml.FullLoader)
        except yaml.YAMLError as exc:
            return exc.args

    @property
    def observable_ids(self):
        return list(self.observables.keys())

    @property
    def station(self) -> Station:
        try:
            return self._station
        except AttributeError:
            return add_item(Station(**self.contents['Station']))

    @property
    def raw_station(self):
        return Station(**self.contents['Station'])

    @station.setter
    def station(self, value):
        self._station = add_item(value)

    @property
    def observables(self) -> [AbstractObservable]:
        return {item['observable_id']: add_item(AbstractObservable(**item))
                for item in self.contents['Observables']}

    @property
    def sensors(self):
        sensors = dict()
        if self.contents['Sensors'] is None:
            return sensors
        for sensor in self.contents['Sensors']:
            relevant_observables = map(
                lambda rel_obs: rel_obs.strip().lstrip().rstrip(),
                sensor.pop('relevant_observables').split(','))
            for observable_id in relevant_observables:
                sensor['abstract_observable_id'] = self.observables[
                    observable_id].id
                sensors[observable_id] = add_item(Sensor(**sensor))
        return sensors

    @property
    def units_of_measurement(self):
        unit = dict()
        for sensor in self.contents['Units of Measurement']:
            relevant_observables = map(
                lambda rel_obs: rel_obs.strip().lstrip().rstrip(),
                sensor.pop('relevant_observables').split(','))
            for observable_id in relevant_observables:
                unit[observable_id] = add_item(UnitOfMeasurement(**sensor))
        return unit



from sqlalchemy import Column, Integer, DateTime, String, ForeignKey

from edam.reader.base import Base


class Observation(Base):
    """
    Represents an observation.

    This ORM class represents an observation. For the time being,
    apart from `value` (in string format) and `timestamp`, we store Foreign
    Keys pointing to the other dimension of the observations. These include,
    the abstract observable, sensor, unit of measurement and station.

    Attributes:
        value: The measurement as string
        timestamp: When the measurement was taken
        junction_id: FK to the corresponding junction table. This tables
            holds further info about the corresponding unit of measurement,
            station and sensor
    """
    __tablename__ = "Observation"
    id = Column(Integer, primary_key=True)
    timestamp = Column(DateTime)
    value = Column(String(60))
    junction_id = Column(Integer, ForeignKey('Junction.id'))

    def __init__(self, timestamp=None, value=None, junction_id=None):
        self.timestamp = timestamp
        self.value = value
        self.junction_id = junction_id

    def __repr__(self):
        return f'<{self.__class__.__name__} with id {self.id!r}>'



import logging
import sys
from logging.handlers import TimedRotatingFileHandler
import os

import yaml
from sqlalchemy.engine.url import URL

ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
home_directory = os.path.join(os.path.expanduser('~'), '.edam')
test_resources = os.path.join(ROOT_DIR, os.pardir, 'tests', 'resources')
settings = os.path.join(home_directory, 'settings.yaml')

with open(settings, 'r') as stream:
    try:
        settings_content = yaml.load(stream, Loader=yaml.BaseLoader)
    except yaml.YAMLError as exc:
        raise exc


def safe_return(section: str, fields: dict) -> dict:
    """
    :param section:
    :param fields:
    :return:
    """
    temp_dict = dict()
    for key, default_value in fields.items():
        try:
            value = settings_content[section][key]
        except KeyError:
            value = default_value
        temp_dict[key] = value

    if section == 'DATABASE':
        if temp_dict['drivername'] in ['sqlite']:
            temp_dict['database'] = os.path.join(
                home_directory, temp_dict['database'])

    return temp_dict


database_fields = {'drivername': 'sqlite', 'database': 'edam.db'}
DATABASE = safe_return('DATABASE', database_fields)

server_fields = {'host': '127.0.0.1', 'port': '5000', 'debug': False}
SERVER = safe_return('SERVER', server_fields)
ogc_sos_fields = {'title': 'EDAM',
                  'description': 'EDAM OGC Sensor Observation Service',
                  'name': 'First name Last name',
                  'provider_name': 'Wageningen University',
                  'voice': '00302541-079-516',
                  'fax': '00302541-079-516',
                  'delivery_point': 'WUR',
                  'city': 'Wageningen',
                  'administration_area': 'Netherlands',
                  'postal_code': '',
                  'country': 'Netherlands'}

OGC_SOS_CONFIGURATION = safe_return('OGC SOS', ogc_sos_fields)
database_url = URL.create(**DATABASE)
database_type = DATABASE['drivername']

FORMATTER = logging.Formatter(
    "%(asctime)s — %(name)s — %(levelname)s — %(message)s")
LOG_FILE = os.path.join(os.path.expanduser('~'), '.edam', "edam.log")
log_level = safe_return('GENERAL', {'loglevel': 'INFO'})
log_level = logging.getLevelName(log_level['loglevel'].upper())


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_file_handler():
    file_handler = TimedRotatingFileHandler(LOG_FILE, when='midnight')
    file_handler.setFormatter(FORMATTER)
    return file_handler


def get_logger(logger_name):
    logger = logging.getLogger(logger_name)
    logger.setLevel(log_level)
    logger.addHandler(get_console_handler())
    logger.addHandler(get_file_handler())
    logger.propagate = False
    return logger



from sqlalchemy import Column, Integer, ForeignKey
from sqlalchemy.orm import relationship

from edam.reader.base import Base


class Junction(Base):
    """
    """
    __tablename__ = "Junction"
    id = Column(Integer, primary_key=True)
    abstract_observable_id = Column(
        Integer, ForeignKey('AbstractObservable.id'))
    observable = relationship("AbstractObservable", back_populates="junctions")

    unit_id = Column(Integer, ForeignKey('UnitOfMeasurement.id'))
    unit = relationship("UnitOfMeasurement", back_populates="junctions")

    station_id = Column(Integer, ForeignKey('Station.id'))
    station = relationship("Station", back_populates="junctions")

    sensor_id = Column(Integer, ForeignKey('Sensor.id'))
    sensor = relationship("Sensor", back_populates="junctions")

    def __init__(self, abstract_observable_id=None,
                 unit_id=None, station_id=None, sensor_id=None):
        self.abstract_observable_id = abstract_observable_id
        self.unit_id = unit_id
        self.station_id = station_id
        self.sensor_id = sensor_id

    def __repr__(self):
        return f'<{self.__class__.__name__} with id {self.id!r}>'



import json

from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

from edam import get_logger
from edam.reader.base import Base
from edam.reader.models.utilities import update_existing

logger = get_logger('edam.reader.models.sensor')


class Sensor(Base):
    __tablename__ = "Sensor"
    """
    Represents a Sensor.

    This ORM class represents a sensor along with its metadata.

    Attributes:
        name: The name of the sensor
        manufacturer: The name of the sensor's manufacturer
        obs_id: The Foreign Key which connects a sensor with 
            an (abstract) observable
        tags: dict attribute which represents any other tags which are not
            explicitly defined
    """
    id = Column(Integer, primary_key=True)
    name = Column(String(60))
    manufacturer = Column(String(60))
    _tags = Column('tags', String(500))

    abstract_observable_id = Column(
        Integer, ForeignKey('AbstractObservable.id'))
    abstract_observable = relationship("AbstractObservable",
                                       back_populates="sensors")

    junctions = relationship("Junction", back_populates="sensor")

    def update(self, new_values: dict):
        update_existing(self, new_values, logger)

    def __init__(self, name: str = None, manufacturer: str = None,
                 abstract_observable_id: int = None,
                 tags: dict = None):
        self.name = name
        self.manufacturer = manufacturer
        self.abstract_observable_id = abstract_observable_id
        self.tags = tags

    @hybrid_property
    def tags(self):
        return json.loads(self._tags)

    @tags.setter
    def tags(self, value):
        self._tags = json.dumps(value)

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} with id {self.id!r}>'



import io
import itertools
import logging
import os
import re
from typing import TYPE_CHECKING, List
from datetime import datetime
import pandas as pd

from edam.reader.database_handler import add_items, add_item
from edam.reader.models.junction import Junction
from edam.reader.models.observation import Observation

logger = logging.getLogger('edam.reader.resolvers.resolver_utilities')

if TYPE_CHECKING:
    from edam.reader.resolvers.resolver import Resolver


def template_matches_file(template_file: str, input_file: str) -> bool:
    return False


def walk_files_in_directory(directory: str) -> List[str]:
    all_files = []
    for root, dirs, files in os.walk(directory):
        for filename in files:
            all_files.append(os.path.join(root, filename))
    return all_files


def store_data_sqlite(resolver: "Resolver"):
    observations = []
    for measurement, dataframe in resolver.timeseries.items():
        abstract_observable_id = \
            resolver.metadata.restructured_metadata[measurement][
                'observable'].id
        sensor_id = resolver.metadata.restructured_metadata[measurement][
            'sensor'].id
        unit_of_measurement_id = \
            resolver.metadata.restructured_metadata[measurement][
                'unit_of_measurement'].id
        station_id = resolver.metadata.station.id
        junction = Junction(abstract_observable_id=abstract_observable_id,
                            unit_id=unit_of_measurement_id,
                            station_id=station_id,
                            sensor_id=sensor_id)
        junction = add_item(junction)  # type: Junction
        for timestamp, value in dataframe.to_dict().items():
            observations.append(Observation(timestamp=timestamp,
                                            value=str(value),
                                            junction_id=junction.id))
    add_items(observations)


def generate_timeseries(resolver: "Resolver") -> [pd.Series]:
    timestamp_columns = list(filter(lambda column: "timestamp." in column,
                                    resolver.template.used_columns))

    contents = resolver.content_as_list[resolver.template.header_line:]
    if contents[0].count(',') == 0:
        contents = '\n'.join(list(
            map(lambda line: re.sub(r'\s+', ',', line.lstrip()).rstrip(
                ',').lstrip(
                ','), contents)))
    else:
        contents = '\n'.join(contents)

    def date_parser(row):
        if "timestamp.dayofyear" in timestamp_columns:
            dt = datetime.strptime(row, "%Y %j")
        else:
            dt = pd.to_datetime(row, errors="coerce")
        return dt

    dataframe_kwargs = {
        'filepath_or_buffer': io.StringIO(contents),
        'names': resolver.template.used_columns,
        'parse_dates': {"timestamp": timestamp_columns},
        'infer_datetime_format': True,
        'na_values': resolver.metadata.station.missing_data,
        'date_parser': date_parser,
        "on_bad_lines": 'warn'
    }
    if resolver.header != '':
        dataframe_kwargs["skiprows"] = [0]

    df = pd.read_csv(**dataframe_kwargs)
    # Drop rows where timestamp was not parsed correctly
    df = df.loc[~df.timestamp.isnull()]
    df.set_index(keys=['timestamp'], inplace=True)
    timeseries = {}
    for variable in resolver.template.variables:
        if variable == "timestamp":
            continue
        timeseries[variable] = df[variable]
        qualifiers = resolver.metadata.station.qualifiers.items()
        for qualifier_name, qualifier in qualifiers:
            if qualifier_name == "missing_data":
                continue

            timeseries[variable] = timeseries[variable].apply(
                lambda x: str(x).rstrip(qualifier))
        if timeseries[variable].dtype is not float:
            timeseries[variable] = timeseries[variable].apply(
                lambda x: float(x))
    return timeseries


def extract_station_from_preamble(resolver: "Resolver"):
    station_dictionary = dict()
    station_dictionary['tags'] = dict()
    station_dictionary['qualifiers'] = dict()
    var_name = r"({{.*?}})"
    for template_line, input_line in itertools.zip_longest(
            resolver.template.preamble.split('\n'),
            resolver.preamble.split('\n')):
        if template_line:
            matches = re.findall(var_name, template_line)

            if matches:
                for match in matches:

                    # input_line = 'Location: 359800E 223800N,
                    # Lat 51.911 Lon -2.584, 67 metres amsl'
                    template_line = template_line.lstrip('')
                    input_line = input_line.lstrip('')
                    # template_line = 'Location: 359800E 223800N,
                    # Lat {{station.latitude}} Lon
                    # {{station.longitude}}, {{station.tags.altitude}}'

                    # new_template_line = ' Lon {{station.longitude}},
                    # {{station.tags.altitude}}'

                    # to_be_replaced = 'Location: 359800E 223800N, Lat'
                    to_be_replaced = template_line.partition(match)[
                        0].strip('\n\r')
                    template_line = template_line.partition(
                        match)[-1].strip('\n\r')

                    # value_of_placeholder =
                    # input_line.partition(template_line.partition('{')[0])[0]
                    if to_be_replaced.strip(' ') == "":
                        value_of_placeholder = input_line
                    else:
                        value_of_placeholder = input_line.replace(
                            to_be_replaced, '').strip(
                            '\n\r')
                    # print(value_of_placeholder)
                    temp_more_curly = template_line.partition('{')[
                        0].strip('\n\r')

                    if temp_more_curly == '':
                        pass
                    else:
                        input_line = value_of_placeholder
                        value_of_placeholder = \
                            value_of_placeholder.partition(temp_more_curly)[
                                0]
                        input_line = input_line.replace(
                            value_of_placeholder, '')

                    # eg ['station', 'latitude'] or ['station', 'tags', 'key']
                    placeholder_var_in_list = match.strip("{}").split('.')
                    # remove 'station', ie first element
                    del placeholder_var_in_list[0]
                    # Now it should be either ['latitude'] or ['tags',
                    # 'key']
                    if placeholder_var_in_list[0] in ['tags', 'qualifiers']:
                        station_dictionary[placeholder_var_in_list[0]][
                            placeholder_var_in_list[-1]] = value_of_placeholder
                    else:
                        station_dictionary[
                            placeholder_var_in_list[0]] = value_of_placeholder

    return station_dictionary



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath('../..'))
sys.setrecursionlimit(1500)

# -- Project information -----------------------------------------------------

project = 'EDAM'
copyright = '2020, Argyrios Samourkasidis'
author = 'Argyrios Samourkasidis'

# The full version, including alpha/beta/rc tags
release = '0.0.1'

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ['sphinx.ext.autodoc',
              'sphinx.ext.napoleon',
              'sphinx.ext.viewcode']

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'alabaster'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
html_theme_options = {
    "show_powered_by": False,
    "github_user": "BigDataWUR",
    "github_repo": "EDAM",
    "github_banner": True,
    "show_related": False,
    "note_bg": "#FFF59C",
}

# html_sidebars = {
#     "index": ["sidebarintro.html", "sourcelink.html", "searchbox.html", "hacks.html"],
#     "**": [
#         "sidebarlogo.html",
#         "localtoc.html",
#         "relations.html",
#         "sourcelink.html",
#         "searchbox.html",
#         "hacks.html",
#     ],
# }

# If true, the index is split into individual pages for each letter.
# html_split_index = False

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# If true, "Created using Sphinx" is shown in the HTML footer. Default is True.
html_show_sphinx = True

# If true, "(C) Copyright ..." is shown in the HTML footer. Default is True.
html_show_copyright = True

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']



from edam.reader.base import Base, engine
from edam.reader.models.junction import Junction
from edam.reader.models.station import Station
from edam.reader.models.observable import AbstractObservable
from edam.reader.models.sensor import Sensor
from edam.reader.models.unit_of_measurement import UnitOfMeasurement
from edam.reader.models.observation import Observation


def recreate_database():
    Base.metadata.drop_all(engine)
    Base.metadata.create_all(bind=engine)



from abc import ABC, abstractmethod

import pandas as pd

from edam import get_logger
from edam.reader.models.metadata import Metadata
from edam.reader.models.template import Template
from edam.reader.resolvers.resolver_utilities import generate_timeseries, \
    extract_station_from_preamble, store_data_sqlite
from edam.utilities.exceptions import TemplateInputHeaderMismatch

logger = get_logger('edam.reader.resolvers.resolver')


class Resolver(ABC):
    def __init__(self, input_uri, template: Template, metadata: Metadata):
        self.template = template
        self.metadata = metadata
        self.input_uri = input_uri
        self.content_as_list = input_uri
        if self.template_matches_input():
            self.complement_stations_from_preamble()
        # self.store_timeseries()

    @property
    def template(self) -> Template:
        return self._template

    @template.setter
    def template(self, value):
        self._template = value

    @property
    def metadata(self) -> Metadata:
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = value

    @property
    def preamble(self) -> str:
        preamble = '\n'.join(self.content_as_list[:self.template.header_line])
        if preamble == '':
            return ''
        return preamble

    @property
    def header(self) -> list:
        template_header_line = self.template.header_line

        header = None
        for index, line in enumerate(self.content_as_list):
            if index == int(template_header_line):
                header = line.strip('\r\n')
                break

        if self.template.header == header:
            return header
        else:
            raise TemplateInputHeaderMismatch(
                f"Template/Input header mismatch.\n"
                f"template: {self.template.header}")

    @property
    def content(self) -> str:
        return '\n'.join(self.content_as_list[self.template.header_line:])

    @property
    def timeseries(self) -> [pd.DataFrame]:
        return generate_timeseries(self)

    def complement_stations_from_preamble(self):
        if self.template.preamble and self.preamble:
            station_dictionary = extract_station_from_preamble(self)
            station = self.metadata.raw_station
            station.update(station_dictionary)
            self.metadata.station = station

    def template_matches_input(self) -> bool:
        if self.template.header == self.header:
            return True
        return False

    def store_timeseries(self):
        logger.debug(f"Storing timeseries for {self.metadata.station.name}")
        store_data_sqlite(resolver=self)

    @property
    @abstractmethod
    def content_as_list(self) -> []:
        pass

    @content_as_list.setter
    @abstractmethod
    def content_as_list(self, value):
        pass



import requests

from edam.reader.resolvers.resolver import Resolver


class HttpResolver(Resolver):

    @property
    def content_as_list(self):
        return self._content_as_list

    @content_as_list.setter
    def content_as_list(self, file_uri):
        response = requests.get(file_uri)
        self._content_as_list = response.text.split('\r\n')



import calendar
import hashlib
import os
import shutil
from datetime import datetime
from functools import wraps, update_wrapper

import jinja2
from flask import Flask
from flask import make_response
from flask_caching import Cache
from flask_sqlalchemy import SQLAlchemy

from edam import get_logger
from edam.reader.database_handler import get_all
from edam import home_directory
from edam.utilities.reader_utilities import find_templates_in_directory
from edam.viewer import config
from edam.viewer.app.utilities import template_matches_source

# These are needed despite not all being used
from edam.reader.models.station import Station
from edam.reader.models.junction import Junction
from edam.reader.models.template import Template
from edam.reader.models.unit_of_measurement import UnitOfMeasurement
from edam.reader.models.observable import AbstractObservable
from edam.reader.models.sensor import Sensor

logger = get_logger('edam.viewer.app')

app = Flask(__name__,
            static_folder=os.path.join(home_directory, '.viewer/', 'static'),
            static_url_path=os.path.join(home_directory, '.viewer/'))
app.config.from_object(config)

db = SQLAlchemy(app)

# Jinja trim whitespace
app.jinja_env.trim_blocks = True
app.jinja_env.lstrip_blocks = True

# Change the folder of templates and static files
app.jinja_loader = jinja2.FileSystemLoader(
    [home_directory + '/.viewer/templates'])

# Don't sort with jsonify
app.config['JSON_SORT_KEYS'] = False

# Set cache
cache = Cache(app, config={'CACHE_TYPE': 'simple'})
cache.init_app(app)

# secret key for sessions
app.secret_key = "this should be harder to guess"


def stations():
    all_stations = get_all(Station)
    return all_stations


def templates():
    all_templates = find_templates_in_directory()
    return all_templates


def render_data(template, station):
    if template_matches_source(template=template, station=station):
        return station.data_iter(template)
    else:
        return None


def nocache(view):
    @wraps(view)
    def no_cache(*args, **kwargs):
        response = make_response(view(*args, **kwargs))
        response.headers['Last-Modified'] = datetime.now()
        response.headers[
            'Cache-Control'] = 'no-store, no-cache, must-revalidate, ' \
                               'post-check=0, pre-check=0, max-age=0'
        response.headers['Pragma'] = 'no-cache'
        response.headers['Expires'] = '-1'
        return response

    return update_wrapper(no_cache, view)


def hash(anything):
    return '_' + hashlib.md5(anything).hexdigest()


app.jinja_env.globals.update(hash=hash)


def d2s(datetime_object):
    return datetime_object.strftime("%D %H %M")


app.jinja_env.globals.update(d2s=d2s)


def same_timestamp(*args):
    if args:
        return args[0]
    else:
        return None


app.jinja_env.globals.update(same_timestamp=same_timestamp)



from edam.reader.models.station import Station
from edam.reader.models.template import Template


def template_matches_source(template: Template, station: Station):
    if set(station.observable_ids + ['timestamp']) >= set(
            template.observable_ids):
        return True
    return False



import os
from enum import Enum
from os.path import expanduser
from typing import List

import requests

from edam import get_logger
from edam.reader.models.metadata import Metadata
from edam.reader.models.template import Template
from edam.reader.resolvers.file_resolver import FileResolver
from edam.reader.resolvers.http_resolver import HttpResolver
from edam.reader.resolvers.resolver import Resolver
from edam.reader.resolvers.resolver_utilities import walk_files_in_directory
from edam.utilities.exceptions import UrlInputParameterDoesNotExist, \
    InputParameterDoesNotExist, TemplateDoesNotExist, \
    MetadataFileDoesNotExist, TemplateInputHeaderMismatch

logger = get_logger('edam.reader.resolvers.resolver_factory')


class InputType(Enum):
    FILE = 1
    FOLDER = 2
    DATABASE = 3
    HTTP = 4


class ResolverFactory:
    __input_type = None

    def __init__(self, input_uri, template, metadata_file, **kwargs):
        self.input_uri = input_uri
        self.template = template
        self.metadata_file = metadata_file
        self.metadata = self.metadata_file
        try:
            self.var = kwargs['var']
        except KeyError:
            pass

    @property
    def input_uri(self):
        return self._input_uri

    @input_uri.setter
    def input_uri(self, value: str):
        if value.startswith("http://") or value.startswith("https://"):
            self.__input_type = InputType.HTTP
            self._input_uri = value
        elif value.__contains__('db'):
            # TODO: implement this check
            pass
        else:
            if os.path.isfile(value):
                # user gave full path
                self._input_uri = os.path.abspath(value)
                self.__input_type = InputType.FILE
            elif os.path.isfile(
                    os.path.join(expanduser("~"), '.edam', 'inputs', value)):
                # user gave relative path inside the ~/edam/input directory
                self._input_uri = os.path.join(expanduser("~"), '.edam',
                                               'inputs', value)
                self.__input_type = InputType.FILE
            elif os.path.isdir(value):
                self._input_uri = os.path.abspath(value)
                self.__input_type = InputType.FOLDER
            elif os.path.isdir(
                    os.path.join(expanduser("~"), '.edam', 'inputs', value)):
                self._input_uri = os.path.join(expanduser("~"), '.edam',
                                               'inputs', value)
                self.__input_type = InputType.FOLDER
            else:
                raise InputParameterDoesNotExist(
                    "{input_uri} does not exist".format(input_uri=value))

    @property
    def template(self):
        return self._template

    @template.setter
    def template(self, value):
        if value:
            if os.path.isfile(value):
                # user gave full path
                self._template = Template(path=os.path.abspath(value))
            elif os.path.isfile(
                    os.path.join(expanduser("~"), '.edam', 'templates', value)):
                # user gave relative path inside the ~/edam/templates directory
                self._template = Template(
                    path=os.path.abspath(
                        os.path.join(expanduser("~"), '.edam', 'templates',
                                     value)))
            else:
                raise TemplateDoesNotExist(f"{value} does not exist")
        else:
            raise TemplateDoesNotExist("No template was given")

    @property
    def metadata_file(self):
        return self._metadata_file

    @metadata_file.setter
    def metadata_file(self, value):
        if os.path.isfile(value):
            # user gave full path
            self._metadata_file = os.path.abspath(value)
        elif os.path.isfile(
                os.path.join(expanduser("~"), '.edam', 'metadata', value)):
            # user gave relative path inside the ~/edam/templates directory
            self._metadata_file = os.path.abspath(os.path.join(expanduser("~"),
                                                               '.edam',
                                                               'metadata',
                                                               value))
        else:
            raise MetadataFileDoesNotExist(f"{value} does not exist")

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value):
        self._metadata = Metadata(path=value)

    @property
    def resolver(self) -> List[Resolver]:
        if self.__input_type is InputType.FILE:
            return [FileResolver(template=self.template,
                                 metadata=self.metadata,
                                 input_uri=self.input_uri)]
        elif self.__input_type is InputType.HTTP:
            resolvers = []
            try:
                var = self.__getattribute__('var')
            except AttributeError:
                var = None
            if var is not None:
                variables = list(map(lambda item: item.lstrip().rstrip(),
                                     var.split(",")))
                for variable in variables:
                    uri = self.input_uri.replace("{_var_}", variable)
                    try:
                        requests.get(uri).raise_for_status()
                        meta = Metadata(path=self.metadata_file)
                        resolvers.append(HttpResolver(template=self.template,
                                                      metadata=meta,
                                                      input_uri=uri))
                    except requests.HTTPError:
                        logger.exception(f"{variable} station does not exist")
                return resolvers
            else:
                return [HttpResolver(template=self.template,
                                     metadata=self.metadata,
                                     input_uri=self.input_uri)]

        elif self.__input_type is InputType.FOLDER:
            resolvers = []
            for file in walk_files_in_directory(self.input_uri):
                try:
                    meta = Metadata(path=self.metadata_file)
                    temp_resolver = FileResolver(template=self.template,
                                                 metadata=meta,
                                                 input_uri=file)
                    if temp_resolver.template_matches_input():
                        resolvers.append(temp_resolver)
                except TemplateInputHeaderMismatch:
                    logger.warning(
                        f"{self.template.filename} can't parse {file}")
                except UnicodeDecodeError:
                    logger.exception(
                        f"Exception while trying to parse {file} with "
                        f"{self.template.filename}")

            return resolvers



from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship

from edam.reader.base import Base


class AbstractObservable(Base):
    """
    Represents an observable.

    This ORM class represents an observable in its abstract form.
    For example, `temperature` is the abstract observable for `temperature_max`
    or `temp_min`, etc.

    Attributes:
        name: The name of the observable (e.g. Temperature)
        ontology: An ontology for the observable
    """
    __tablename__ = "AbstractObservable"
    id = Column(Integer, primary_key=True)
    observable_id = Column(String(60))
    name = Column(String(60))
    ontology = Column(String(160))

    sensors = relationship("Sensor", back_populates="abstract_observable")

    junctions = relationship("Junction", back_populates="observable")

    def __init__(self, name=None, ontology=None, observable_id=None,
                 **kwargs):
        self.name = name
        self.ontology = ontology
        self.observable_id = observable_id

    def __repr__(self):
        return f'<{self.__class__.__name__} {self.name} with id {self.id!r}>'



import copy
import errno
import fnmatch
import io
import itertools
import json
import logging
import os
import re
from datetime import datetime
from enum import Enum

import numpy as np
import pandas as pd
import records
import requests

# from edam.reader.manage import DatabaseHandler
from edam.reader.models.station import Station
from edam.reader.models import template
from edam.reader.models.template import Template
from edam import database_type, home_directory

logger = logging.getLogger('edam.reader.utilities')


def find_templates_in_directory(directory: str = home_directory) -> [Template]:
    """
        Returns a list of templates located in a given directory.
        Directory defaults to the home_directory(~/.edam)
        :rtype: [Template]
        :param directory: Directory to search for templates
        :return: list of Template objects
    """
    list_of_templates = list()
    for folder, _, files in os.walk(os.path.join(directory, 'templates')):
        for file in files:
            if fnmatch.fnmatch(file, '*.tmpl'):
                path = os.path.join(folder, file)
                filename, _ = os.path.splitext(file)
                template = Template(path=path)
                list_of_templates.append(template)
    return list_of_templates


def remove_template_placeholders_from_string(string, separator="({.*?}+)"):
    """
    This function cleans a **string** according to a given **separator**.

    :param string: The string which will be cleaned
    :param separator: Regex pattern. $param1 will be cleaned according to this param.
    :return: string: Cleaned string
    """

    temp_list = re.split(separator, string, 1)
    # string does not contain separator. Return as is
    if len(temp_list) == 1:
        returning_value = temp_list[0]
    else:
        # It means it has 3 elements.
        # returning_value = temp_list[0] +
        # remove_template_placeholders_from_string(temp_list[2], separator)
        returning_value = temp_list[0]

    return returning_value


def template_matches_input_file(template_file: Template, data_input: Template):
    """
    We are going to work with every line (i.e. list element) individually.
    Specifically, we are going to check each of the lines against input_file files.
    If all lines match to Va given input_file file, it means that we can parse data
    out of it.

    :param template_file: The name of the template (e.g. Yucheng.tmpl)
    :param data_input: Either data_input filename (e.g. Yucheng.met) or string
    with content stored in memory
    :return: True in case match is 100% or False in case match is NOT 100%
    """

    header = template_file.header
    logger.debug("Check lines: %s")
    matching_percentage = int()

    logger.debug("Input file: %s" % data_input)

    for header_item in header:
        if header_item in ":":
            matching_percentage += 1

    # percentage = float(100 * matching_percentage / len(check_lines))
    percentage = 10
    if percentage > 20:
        return True
    else:
        return False


def evaluate_variable_part(variable):
    """
    This function evaluates a variable (check parameter).
    :param variable:{0-n}
    :return: ['{0-n}': [0, 1, 2, ..., n]]
    """
    initial_var = copy.deepcopy(variable)
    variable = variable.strip('{}').split('-')
    if len(variable) > 2 or len(variable) < 2:
        logger.error('Range %s is not correct in the correct format '
                     '{starting_int - ending_int}' % initial_var)
        raise Exception('Range %s is not correct in the correct format '
                        '{starting_int - ending_int}' % initial_var)

    starting_number = variable[0]
    ending_number = variable[1]

    # We need this one to set the argument of "format" function
    # If we are given 01-03, we should return 01, 02, 03
    # and NOT 1, 2, 3

    length = len(starting_number)
    try:
        starting_number = int(starting_number)
        # range() is not inclusive!
        ending_number = int(ending_number) + 1
    except BaseException:
        logger.error(
            "Can't convert variables to integers %s, %s" % (
                starting_number, ending_number))
        raise Exception("Can't convert variables to integers")

    the_range = range(starting_number, ending_number)
    # e.g. format(x,'02d')
    format_argument = '0' + str(length) + 'd'
    temp_list = [format(x, format_argument) for x in the_range]
    temp_as_a_dict = dict()
    temp_as_a_dict[initial_var] = temp_list

    return temp_as_a_dict


def generate_uri(uri: str, static_variables=None):
    """
    This function generates a URI derivatives out of a URI which contains a
    variable part (i.e. {number-number}). Variable parts are identified
    with regex. A URI may contain more than one variable parts
    (e.g. https://example.com/{01-02}-{2010-2011}).
    Function should parse them both and generate variable parts **seperately**.
    For the above example, output is:
    - https://example.com/01-2010
    - https://example.com/02-2010
    - https://example.com/01-2011
    - https://example.com/02-2011

    Attention: There may be cases in which we have the **very same**
    variable part more than one times. All these parts should be generated
    **all-together** (and not seperately). Consider the following example:
    https://www.example.com/{2016-2017}{01-02}/IDCJDW2006.{2016-2017}{01-02}
    We can identify to "clusters": {2016-2017} and {01-02} occuring twice each.

    For the above example, output is:
    - https://www.example.com/201601/IDCJDW2006.201601
    - https://www.example.com/201701/IDCJDW2006.201701
    - https://www.example.com/201602/IDCJDW2006.201602
    - https://www.example.com/201702/IDCJDW2006.201702

    This version supports only for numerical operations.

    It also exports a list of uris based on a static input_file.
    A URI contains a placeholder (i.e. https://example.com/{$var}).
    User defines also a set of vars ("one, two, three, four").

    For the above example, output is:
    - https://example.com/one
    - https://example.com/two
    - https://example.com/three
    - https://example.com/four

    :param uri: URI with variable member as indicated by \
    user (e.g. https://example.com/{01-09})
    :param static_variables:
    :return: A list with all resulting URIs. In case URI does \
    not contain variable part (i.e. {01-09}), it returns one uri in a list
    """
    # variable_regex is more generic. It matches both {01-09} AND {$var}
    # (i.e. static_variable_regex)
    variable_regex = r"({.*?})"
    static_variable_regex = r"({\$var})"
    match = re.search(variable_regex, uri)
    uris_in_a_list = list()
    if match:
        if re.search(static_variable_regex, uri):
            try:
                # We are in the case we have static variables
                static_variables_in_a_list = static_variables.split(',')
                for static_var in static_variables_in_a_list:
                    iteration_uri = copy.deepcopy(uri)
                    iteration_uri = iteration_uri.replace(
                        "{$var}", static_var.strip())
                    uris_in_a_list.append(iteration_uri)
            except AttributeError:
                logger.error(
                    '--extra parameter was not given (Station variables). '
                    'generate_uri()')
                raise AttributeError(
                    '--extra parameter (static vars) was empty')
        else:
            # We want all unique variable parts. See documentation about
            # clusters!
            variable_parts_in_a_list = list(
                set(re.findall(variable_regex, uri)))
            evaluated_variable_parts = list(
                map(evaluate_variable_part, variable_parts_in_a_list))

            to_be_replaced = list()
            to_replace_with = list()
            for dict_pair in evaluated_variable_parts:
                key, value = dict_pair.popitem()
                to_be_replaced.append(key)
                to_replace_with.append(value)

            to_replace_with = list(itertools.product(*to_replace_with))

            for item in to_replace_with:
                iteration_uri = copy.deepcopy(uri)
                for count, value in enumerate(item):
                    iteration_uri = iteration_uri.replace(
                        to_be_replaced[count], value)
                uris_in_a_list.append(iteration_uri)
    else:
        uris_in_a_list.append(uri)

    return uris_in_a_list


if __name__ == "__main__":
    pass



import re

start_if = re.compile(r"{%\s?if.*%}")
end_if = re.compile(r"{%\s?endif\s?%}")
variable_name = re.compile(r"{{(.*?)}}")
for_loop_variables = re.compile(r"{%\s?for (.*?) in (.*?)%}")
end_for_loop = re.compile(r"{%endfor%}")
template_file_header = r"((.*\n){1}){%\s?for.*%}"
# This var corresponds to observables which value is located in more than
# one columns
# In our templating language they are represented as:
# {{observable.value[0]}},{{observable.value[1]}}, etc.
# With that regex we are going to determine the existence of those observables
multicolumn_values_bom = re.compile(r"{{.*?\..*?(\[\d\])*}}")

# For the following regex, line has to be split to ',' and then
# each individual var to be parsed with var_name
var_for_line = re.compile(r"{%\s?for .*? in .*?%}(.*){%\s?endfor\s?%}",
                          re.DOTALL)
var_parse_header = re.compile(r"(.*?)\n*{%\s?for .*? in .*?%}")

# var_anything_but_value = re.compile(r"{{.*?.value(.*?)}}")
# var_anything_but_value = re.compile(r"{{.+.value(\[.+\]).?.?}}")
var_for_value_timestamp = re.compile(r"{{.+.(timestamp.?)}}")

resample_function = re.compile(r"station\.resampled\(\s*(\S+)\s*,\s*(\S+)\)")



import click

from edam.reader.models.database import recreate_database
from edam.reader.resolvers.resolver_factory import ResolverFactory
from edam import SERVER
from edam.utilities.decorators import timer
from edam.viewer.app.views import app


@click.command()
@click.option('--input', 'input_file', required=True, help='input string')
@click.option('--template', required=True,
              help='template file to parse data with')
@click.option('--metadata', required=True,
              help='configuration file to annotate data with')
@click.option('--var', required=False, default="",
              help='Extra variables for URI generation')
@click.option('--drop', required=False, default='no',
              type=click.Choice(['yes', 'no']),
              help="Whether to drop stored data or not")
@timer
def cli(input_file, template, metadata, var, drop):
    if drop == "yes":
        recreate_database()

    factory = ResolverFactory(input_uri=input_file,
                              template=template,
                              metadata_file=metadata,
                              var=var)
    resolvers = factory.resolver
    # Workflow
    for resolver in resolvers:
        if resolver.template_matches_input():
            dataframes = resolver.timeseries
            resolver.store_timeseries()


def run():
    app.jinja_env.auto_reload = True
    app.config['TEMPLATES_AUTO_RELOAD'] = True
    app.run(host=SERVER['host'], port=SERVER['port'], debug=SERVER['debug'])


if __name__ == '__main__':
    cli()
